@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico
-   Copyright (C) 2007-2008, 2010, 2012 Sergey Poznyakoff
+   Copyright (C) 2007-2008, 2010, 2012, 2016-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1540,9 +1540,19 @@ utf8_strlen(const char *s)
 static int
 utf8_iter0(struct utf8_iterator *itr)
 {
-    size_t n = utf8_char_width(itr->curptr);
-    if (n == 0)
+    size_t n;
+
+    if (itr->length == 0) {
+	itr->end = 1;
 	return 1;
+    }
+    n = utf8_char_width(itr->curptr);
+    if (n > itr->length) {
+	itr->end = 1;
+	itr->err = 1;
+	return 1;
+    }
+    
     itr->curwidth = n;
     return 0;
 }
@@ -1550,23 +1560,40 @@ utf8_iter0(struct utf8_iterator *itr)
 int
 utf8_iter_end_p(struct utf8_iterator *itr)
 {
-    return *itr->curptr == 0;
+    return itr->end;
+}
+
+int
+utf8_iter_err_p(struct utf8_iterator *itr)
+{
+    return itr->end;
+}
+
+int
+utf8_iter_init(struct utf8_iterator *itr, char *ptr, size_t size)
+{
+    itr->string = ptr;
+    itr->curptr = ptr;
+    itr->length = size;
+    itr->curwidth = 0;
+    itr->err = 0;
+    itr->end = 0;
+    return utf8_iter0(itr);
 }
 
 int
 utf8_iter_first(struct utf8_iterator *itr, char *ptr)
 {
-    itr->string = ptr;
-    itr->curptr = ptr;
-    return utf8_iter0(itr);
+    return utf8_iter_init(itr, ptr, strlen(ptr));
 }
 
 int
 utf8_iter_next(struct utf8_iterator *itr)
 {
-    if (*itr->curptr == 0)
+    if (utf8_iter_end_p(itr))
 	return -1;
     itr->curptr += itr->curwidth;
+    itr->length -= itr->curwidth;
     return utf8_iter0(itr);
 }
 
@@ -1590,9 +1617,11 @@ utf8_wctomb(char *r, unsigned int wc)
 	count = 5;
     else if (wc <= 0x7fffffff)
 	count = 6;
-    else
+    else {
+	errno = EILSEQ;
 	return -1;
-
+    }
+    
     switch (count) {
 	/* Note: code falls through cases! */
     case 6:
@@ -1746,77 +1775,88 @@ urf8_symcasecmp(char *a, char *b)
 	return 1;
     return 0;
 }
-
+
 int
-utf8_strcasecmp(char *a, char *b)
+utf8_compare(char const *a, char const *b,
+	     int ci, size_t maxlen, int (*wcsel)(unsigned))
 {
     int alen, blen;
+    size_t an = 0, bn = 0;
+    unsigned wa, wb;
 
-    for (; *a; a += alen, b += blen) {
-	unsigned wa, wb;
-
-	if (*b == 0)
-	    return 1;
-
-	alen = utf8_char_width(a);
-	if (alen == 0)
-	    return -1;
-	utf8_mbtowc(&wa, a, alen);
-	blen = utf8_char_width(b);
-	if (blen == 0)
-	    return 1;
-	utf8_mbtowc(&wb, b, blen);
-	wa = utf8_wc_toupper(wa);
-	wb = utf8_wc_toupper(wb);
-	if (wa < wb)
-	    return -1;
-	if (wa > wb)
-	    return 1;
-
-    }
-    if (*b)
-	return -1;
-    return 0;
-}
-
-int
-utf8_strncasecmp(char *a, char *b, size_t maxlen)
-{
-    int alen, blen;
-    unsigned asz = 0, bsz = 0;
-
-    while (asz < maxlen) {
-	unsigned wa, wb;
-
+    while (1) {
+	if (maxlen != 0 && an == maxlen)
+	    return 0;
 	if (*a == 0)
-	    return (*b == 0) ? 0 : -1;
-
-	if (*b == 0)
-	    return 1;
-
+	    break;
+	
 	alen = utf8_char_width(a);
 	if (alen == 0)
 	    return -1;
+	
 	utf8_mbtowc(&wa, a, alen);
+	a += alen;
+	an++;
+
+	if (!wcsel || wcsel(wa)) {
+	    if (*b == 0)
+		return 1;
+	    while (*b) {
+		if (maxlen != 0 && bn == maxlen)
+		    return 0;
+		blen = utf8_char_width(b);
+		if (blen == 0)
+		    return 1;
+		utf8_mbtowc(&wb, b, blen);
+		b += blen;
+		bn++;
+		
+		if (!wcsel || wcsel(wb)) {
+		    if (ci == case_insensitive) {
+			wa = utf8_wc_toupper(wa);
+			wb = utf8_wc_toupper(wb);
+		    }
+		    if (wa < wb)
+			return -1;
+		    if (wa > wb)
+			return 1;
+		    break;
+		}
+	    }
+	}
+    }
+
+    while (*b) {
 	blen = utf8_char_width(b);
 	if (blen == 0)
 	    return 1;
 	utf8_mbtowc(&wb, b, blen);
-	wa = utf8_wc_toupper(wa);
-	wb = utf8_wc_toupper(wb);
-	if (wa < wb)
-	    return -1;
-	if (wa > wb)
-	    return 1;
-	a += alen;
 	b += blen;
-	asz ++;
-	bsz ++;
+	if (!wcsel || wcsel(wb))
+	    return -1;
     }
+    
     return 0;
 }
 
+int
+utf8_strcmp(char const *a, char const *b)
+{
+    return utf8_compare(a, b, case_sensitive, 0, NULL);
+}
 
+int
+utf8_strcasecmp(char const *a, char const *b)
+{
+    return utf8_compare(a, b, case_insensitive, 0, NULL);
+}
+
+int
+utf8_strncasecmp(char const *a, char const *b, size_t maxlen)
+{
+    return utf8_compare(a, b, case_insensitive, maxlen, NULL);
+}
+
 unsigned
 utf8_wc_toupper(unsigned wc)
 {
@@ -2000,34 +2040,117 @@ utf8_wc_strchr_ci(const unsigned *str, unsigned chr)
 	    return str;
     return NULL;
 }
-    
-const unsigned *
-utf8_wc_strstr(const unsigned *haystack, const unsigned *needle)
+
+#define INIT ((size_t) -1)
+
+/* Compute the table of tagged border lengths for use in the
+   Knuth-Morris-Pratt algorithm.
+*/
+static size_t *
+kmp_next(const unsigned *text, size_t text_len,
+	 const unsigned *pattern, size_t pattern_len)
 {
-    unsigned first = needle[0];
+    size_t *nextab;
+    size_t i, j;
 
-    /* Is needle empty?  */
-    if (first == 0)
-	return haystack;
+    if (pattern_len + 1 == 0) {
+	errno = ERANGE;
+	return NULL;
+    }
+    
+    nextab = calloc(pattern_len + 1, sizeof nextab[0]);
+    if (!nextab)
+	return NULL;
+    
+    i = 0;
+    j = INIT;
+    nextab[0] = j;
+    while (i < pattern_len) {
+	while (j != INIT && pattern[i] != text[j])
+	    j = nextab[j];
+	i++;
+	j++;
+	if (pattern[i] == pattern[j])
+	    nextab[i] = nextab[j];
+	else
+	    nextab[i] = j;
+    }
 
-    /* Is needle nearly empty?  */
-    if (needle[1] == 0)
-	return utf8_wc_strchr(haystack, first);
-    for (; *haystack; haystack++)
-	if (*haystack == first) {
-	    /* Compare with needle's remaining units.  */
-	    const unsigned *hptr = haystack + 1;
-	    const unsigned *nptr = needle + 1;
-	    for (;;) {
-		if (*hptr != *nptr)
-		    break;
-		hptr++;
-		nptr++;
-		if (*nptr == 0)
-		    return haystack;
+    return nextab;
+}
+
+/* Find the first occurrence of PATTERN in TEXT.
+   Return value:
+     strpat_found     - Pattern was found.  Unless RETURN_OFFSET is NULL, the
+                        offset of the pattern is stored in the memory location
+		        it points to.
+     strpat_not_found - Pattern was not found.
+     strpat_error     - An error occurred.  The errno variable is set to
+                        ERANGE if pattern is too long, or to ENOMEM if unable
+			to allocate memory.
+
+   This implementation uses Knuth-Morris-Pratt algorithm.  If low on memory,
+   it falls back to brute force algorithm.
+*/
+unsigned const *
+utf8_wc_strstr(const unsigned *text, const unsigned *pattern)
+{
+    size_t i, j;
+    unsigned const *result;
+    size_t text_len = utf8_wc_strlen(text);
+    size_t pattern_len = utf8_wc_strlen(pattern);
+    size_t *nextab;
+
+    /* Handle corner cases */
+    if (pattern_len > text_len)
+	return NULL;
+    else if (pattern_len == 1)
+	return utf8_wc_strchr(text, pattern[0]);
+    else if (pattern_len == text_len) {
+	if (utf8_wc_strcmp(text, pattern) == 0)
+	    return text;
+	else
+	    return NULL;
+    }
+
+    /* Uses Knuth-Morris-Pratt algorithm for the general case */
+    nextab = kmp_next(text, text_len, pattern, pattern_len);
+    if (!nextab) {
+	/* Fall back to brute force approach */
+	unsigned first = pattern[0];
+	for (; *text; text++)
+	    if (*text == first) {
+		/* Compare with pattern's remaining units.  */
+		const unsigned *text_ptr = text + 1;
+		const unsigned *pat_ptr = pattern + 1;
+		for (;;) {
+		    if (*text_ptr != *pat_ptr)
+			break;
+		    text_ptr++;
+		    pat_ptr++;
+		    if (*pat_ptr == 0)
+			return text;
+		}
 	    }
+	return NULL;
+    }
+
+    i = j = 0;
+    result = NULL;
+    while (j < text_len) {
+	while (i != INIT && pattern[i] != text[j])
+	    i = nextab[i];
+	i++;
+	j++;
+	if (i >= pattern_len) {
+	    result = text + j - i;
+	    break;
 	}
-    return NULL;
+    }
+
+    free(nextab);
+
+    return result;
 }
     
 unsigned *
@@ -2062,10 +2185,10 @@ utf8_wc_to_mbstr(const unsigned *wordbuf, size_t wordlen, char **sptr)
 
     wbc = 0;
     for (i = 0; i < wordlen; i++) {
-	char r[4];
+	char r[6];
 	int rc = utf8_wctomb(r, wordbuf[i]);
 	if (rc <= 0)
-	    return rc;
+	    return -1;
 	wbc += rc;
     }
 
@@ -2077,10 +2200,14 @@ utf8_wc_to_mbstr(const unsigned *wordbuf, size_t wordlen, char **sptr)
 
     wbc = 0;
     for (i = 0; i < wordlen; i++) {
-	char r[4];
+	char r[6];
 	int rc = utf8_wctomb(r, wordbuf[i]);
-	if (rc <= 0)
-	    return rc;
+	if (rc <= 0) {
+	    int ec = errno;
+	    free(s);
+	    errno = ec;
+	    return -1;
+	}
 	memcpy(s + wbc, r, rc);
 	wbc += rc;
     }
@@ -2115,7 +2242,9 @@ utf8_mbstr_to_wc(const char *str, unsigned **wptr, size_t *plen)
     for (i = 0, len = strlen(str); len; i++) {
 	int rc = utf8_mbtowc(w + i, str, len);
 	if (rc <= 0) {
+	    int ec = errno;
 	    free(w);
+	    errno = ec;
 	    return -1;
 	}
 	str += rc;
@@ -2183,3 +2312,29 @@ utf8_quote(const char *str, char **sptr)
     }
     return rc;
 }
+
+int
+utf8_table_check(unsigned wc,
+		 unsigned const *start, int const *count, size_t size)
+{
+    const unsigned *l = start;
+    const unsigned *r = start + size;
+    const unsigned *s = NULL;
+
+    if (wc == 0)
+	return 0;
+
+    while (l < r) {
+	s = l + ((r - l) >> 1);
+
+	if (*s <= wc) {
+	    l = s + 1;
+	} else {
+	    r = s;
+	}
+    }
+    
+    --l;
+    
+    return (wc < l[0] + count[l - start]);
+}
